/****************************************************************************
**
** Copyleft (C) 2015 Dominique Laporte <laporte.educnat@gmail.com>
**
** This file is part of ecoserre
**
** GNU Lesser General Public License Usage
** Alternatively, this file may be used under the terms of the GNU Lesser
** General Public License version 2.1 or version 3 as published by the Free
** Software Foundation and appearing in the file LICENSE.LGPLv21 and
** LICENSE.LGPLv3 included in the packaging of this file. Please review the
** following information to ensure the GNU Lesser General Public License
** requirements will be met: https://www.gnu.org/licenses/lgpl.html and
** http://www.gnu.org/licenses/old-licenses/lgpl-2.1.html.
**
****************************************************************************/

#include <SoftwareSerial.h>

const int RX = 2;    // pin Shield Arduino pour reception
const int TX = 3;    // pin Shield Arduino pour transmission

const char *PANID = "";      // adresse reseau (laisser à vide si XBee absent)
//const char *PANID = "1111";      // adresse reseau
const char *MYID  = "1001";      // adresse du XBee
const char *DLID  = "1002";      // adresse de destination
const char *CHID  = "10";        // n° de canal

//---- paramétrage du temps entre chaque lecture des entrées numériques et analogiques ----//
const int  refresh = 5;  // l'unité est en secondes

//---- prise en compte de la somme de controle ----//
const bool  checksum = false;

//---- paramétrage pour la liaison serie ----//
const int  baudRate = 9600; // vitesse de communication

// variables réservées, ne pas modifier
SoftwareSerial xbeeSerial(TX, RX);     // pin 2 = RX Arduino, pin3 = TX Arduino
unsigned long temps = millis();   // variable qui stocke la mesure du temps
boolean mode_auto = true;  // mode auto ou manuel pour la régulation

unsigned int IOnumber;  // nombre d'actuateurs
const unsigned int MAXIO = 20;
struct _actuateur {   // definition actuateurs
  String ident;   // identification de l'actuateur (cf base de donnee cote serveur)
  boolean type;   // numerique (true) ou analogique
  boolean io;     // entree (true) ou sortie (false)
  int pin;        // numero broche arduino 
  int low;        // valeur inf pour actionneur (cf fonction regulation)
  int high;       // valeur sup pour actionneur
  } actuateur[MAXIO];

//==== la régulation : fonction à compléter ====//
void regulation()
{
  //-----------------------------------------//
  // exemple...
  int in  = getIndex("lum");    // index du capteur (luxmètre)
  int out = getIndex("lamp");   // index de l'actionneur (lampe pour éclairage)
  if ( analogRead(actuateur[in].pin) < actuateur[in].low )
    digitalWrite(actuateur[out].pin, HIGH);
  else if ( analogRead(actuateur[in].pin) > actuateur[in].high )
    digitalWrite(actuateur[out].pin, LOW);
  //-----------------------------------------//
}

//==== initialisation du programme : s'exécute une seule fois ====//
void setup()
{
   // on initialise la liaison série
  Serial.begin(baudRate);
  
  if ( strlen(PANID) ) {
    xbeeSerial.begin(baudRate);        // Liaison série du module XBEE
    
    Serial.println("XBee init");
    XBEEinit();
    }

  // initialisation structures actuateurs
  IOnumber = _init();
  setActuators( IOnumber );
   
  Serial.println("CMDE:<A|D|M|S>:<pin>:<valeur1>:<valeur2>#");
}

//==== programme proncipal : boucle en continu ====//
void loop()
{
  // la fonction de régulation
  if ( mode_auto )
    regulation();

  // envoie des mesures
  if ( (millis() - temps) > refresh * 1000 ) {
    //==== lecture des mesures ====
    String datalog("DATA;");        // 1er champ : entête
    datalog = datalog + String(MYID) + String(";");   // 2ème champ : ID iot

    for (unsigned int i(0); i < IOnumber; i++) {
      datalog = datalog + actuateur[i].ident + String("=");
      if ( actuateur[i].io )
        datalog = datalog + (actuateur[i].type ? String(digitalRead(actuateur[i].pin)) : String(analogRead(actuateur[i].pin)));
      if ( i < IOnumber - 1 )
        datalog = datalog + String(";");
      }

    // calcul du checksum
    if ( checksum )
      datalog = datalog + String(";") + String(setCheckSum(datalog.c_str()));

    // transmission des mesures
    // enregistrement délimité par saut de ligne
    if ( strlen(PANID) )
      xbeeSerial.println(datalog);
    else
      Serial.println(datalog);
      
    temps = millis(); //on stocke la nouvelle heure
    }
}

//==== initialisation des actuateurs ====//
const int _init()
{
  char *ptr, incomingBytes[512] = "";   // for incoming serial data

    do {      
      if ( strlen(PANID) )
        xbeeSerial.println("INIT");
      else
        Serial.println("INIT");

      // read the incoming byte:
      readByte(incomingBytes, 511);

      if ( (ptr = strstr(&incomingBytes[4], "ACK:")) == NULL )
        ptr = incomingBytes;
      //Serial.print("--1");Serial.println(incomingBytes);
      //Serial.print("--2");Serial.println(ptr);

      // attente ACKwnoledge
      } while ( strncmp(incomingBytes, "ACK:", 4) );
              
  Serial.println("OK");
  
  return parseConfigData(ptr + 4);
}

//---- lecture port série
void readByte(char *buffer, const int _size)
{
    if ( dataAvailable() ) {
      int  i(0);
      int  c = strlen(PANID)
        ? xbeeSerial.read()
        : Serial.read() ;      

      // terminateur caractère #
      while ( c != '#' && i < _size ) {
        if ( c != -1 )
          buffer[i++] = (char) c;

        c = strlen(PANID)
          ? xbeeSerial.read()
          : Serial.read() ;
        }
      buffer[i] = 0;

      // vide le buffer
      if ( strlen(PANID) )
        xbeeSerial.flush();
      else
        Serial.flush();
      }
}

//==== détection de nouvelles données proviennant de la liaison RX matériel ====
void serialEvent()
{
  char incomingBytes[512] = "";   // for incoming serial data

  // read the incoming byte:
  readByte(incomingBytes, 511);

  // traitement de la commande
  if ( !strncmp(incomingBytes, "CMDE:", 5) )
    parseCommandData(&incomingBytes[5]);
}

// Get the number of bytes (characters) available for reading from the serial port
const int dataAvailable()
{
  if ( strlen(PANID) )
    return xbeeSerial.available();

  return Serial.available();
}

// traitement des données reçues
void parseCommandData(const char *str)
{
  Serial.print("RX DATA: "); Serial.println(str);
  
  //== format : <id:type E/S:n° broche:valeur1:valeur2> ==//
  String  string = str;
  
  int _start = string.indexOf(':');
  int _stop  = string.indexOf(':', _start+1);
  String type = string.substring(_start+1, _stop);
  
  _start = _stop;
  _stop  = string.indexOf(':', _stop+1);
  int pin = atoi(string.substring(_start+1, _stop).c_str());
  
  _start = _stop;
  _stop  = string.indexOf(':', _stop+1);
  int param_1 = atoi(string.substring(_start+1, _stop).c_str());
  
  _start = _stop;
  _stop  = string.indexOf(':', _stop+1);
  int param_2 = atoi(string.substring(_start+1, _stop).c_str());

  switch ( type[0] ) {
    case 'M': // mode auto ou manuel
      mode_auto = (boolean) param_1;
      break;
    case 'S': // demande de synchronisation
      IOnumber = _init(); // définition des entrées/sorties numériques
      setActuators( IOnumber );
      break;
    case 'D': // sortie numérique
      digitalWrite(pin, param_1);
      break;
    case 'A': // sortie analogique
    case 'P': // sortie PWM
      if ( param_2 )  // sens moteur
        digitalWrite(abs(param_2), param_2 > 0 ? HIGH : LOW);      // commande du pont en H    
      analogWrite(pin, param_1);     //set pwm control, 0 for stop, and 255 for maximum speed
      break;
    default:
      break;
    }
}

// traitement des données reçues
const int parseConfigData(const char *str)
{
  Serial.print("RX CFG: "); Serial.println(str);
    
  //== format : <ident:type:I/O:pin:min:max> ==//
  // séparateur ;
  String  string = str;
  int _start = -1;
  int _stop  = string.indexOf(';', _start+1);
  int i(0);
  
  do {
    String parse = string.substring(_start+1, _stop);
    
    int _first = 0;
    int _next  = parse.indexOf(':', _first+1);
    actuateur[i].ident = parse.substring(_first, _next);
        
    _first = _next;
    _next  = parse.indexOf(':', _next+1);
    actuateur[i].type = (boolean) (parse.substring(_first+1, _next) == "D");
        
    _first = _next;
    _next  = parse.indexOf(':', _next+1);
    actuateur[i].io = (boolean) (parse.substring(_first+1, _next) == "I");
    
    _first = _next;
    _next  = parse.indexOf(':', _next+1);
    actuateur[i].pin = atoi(parse.substring(_first+1, _next).c_str());
    
    _first = _next;
    _next  = parse.indexOf(':', _next+1);
    actuateur[i].low = atoi(parse.substring(_first+1, _next).c_str());
    
    _first = _next;
    _next  = parse.indexOf(':', _next+1);
    actuateur[i++].high = atoi(parse.substring(_first+1, _next).c_str());
  
    _start = _stop;
    _stop  = (_stop != -1 ) ? string.indexOf(';', _stop+1) : -1;
    
    /*
    Serial.print("-------------");
    Serial.print(i); Serial.print(" ");
    Serial.print(_start); Serial.print(" ");
    Serial.print(_stop); Serial.print(" ");
    Serial.println();
    */
    } while ( _stop != _start && i < MAXIO );

  return i;
}

// recherche l'actuateur selon son ident
const int getIndex(const char *ident)
{
  for (unsigned int i(0); i < IOnumber; i++)
    if ( actuateur[i].ident == ident )
      return i;

  return -1;
}

const int setCheckSum(const char *frame)
{
  const char  *ptr = frame;
  int  sum = 0;

  while ( *ptr )
    sum += *ptr++;

  return 0xFF - (sum & 0xFF);
}

//==== initialisation XBee ====
void setActuators(const int _IOnumber)
{
  if ( _IOnumber ) {
    Serial.print("#initialisation : "); Serial.println(_IOnumber);

    for (int i(0); i < _IOnumber; i++)
      if ( actuateur[i].type ) {
        pinMode(actuateur[i].pin, actuateur[i].io ? INPUT : OUTPUT);
        Serial.print(actuateur[i].ident);
        Serial.print(" #"); Serial.print(actuateur[i].pin);
        Serial.println(actuateur[i].io ? " INPUT" : " OUTPUT");
        }
  }
}
        
//==== initialisation XBee ====
void XBEEinit()
{
  setXBEE("+++");

  // attente réponse OK de la commande AT
  char thisByte = 0;
  while ( thisByte != '\r' )
    if ( xbeeSerial.available() > 0 ) {
      thisByte = xbeeSerial.read();
      Serial.print(thisByte);
      }
  Serial.println("");

  char  buffer[11];
  setXBEE("ATRE\r");        // configuration usine par defaut
  setXBEE(strcat(strcat(strcpy(buffer,"ATID"), PANID), "\r"));    // adresse du reseau PAN
  setXBEE(strcat(strcat(strcpy(buffer,"ATMY"), MYID), "\r"));    // set my address using 16-bit addressing
  setXBEE(strcat(strcat(strcpy(buffer,"ATDL"), DLID), "\r"));    // adresse de destination sur 16 bits  
  setXBEE(strcat(strcat(strcpy(buffer,"ATCH"), CHID), "\r"));    // canal 3
  setXBEE("ATDH0\r");       // pas de partie haute, adresse 16 bits
  setXBEE("ATCE0\r");       // end device
  setXBEE("ATWR\r");        // enregistrement
  setXBEE("ATCN\r");        // put the radio in data mode
  
  // vide le tampon d'entrée
  xbeeSerial.flush();
  Serial.println("ready");
}

void setXBEE(char *buffer)
{ 
  Serial.println(buffer);
  Serial.print("sent: "); Serial.println(xbeeSerial.print(buffer));
  delay(100);
}

